from .facet import AttrIntFacetFilter
from .facet import AttrBoolFacetFilter
from .facet import AttrRangeFacetFilter
from .simple import AttrBoolSimpleFilter
from .simple import AttrRangeSimpleFilter
from .simple import AttrIntSimpleFilter


__all__ = [
    'AttrBoolSimpleFilter',
    'AttrBoolFacetFilter',
    'AttrIntFacetFilter',
    'AttrIntSimpleFilter',
    'AttrRangeSimpleFilter',
    'AttrRangeFacetFilter',
]

__version__ = '0.1.4'



import typing as t

from elasticmagic.ext.queryfilter.queryfilter import BaseFilterResult


T = t.TypeVar('T')
TMinValue = t.Union[int, float, None]
TMaxValue = t.Union[int, float, None]


class AttrFacetValue(t.Generic[T]):
    def __init__(
            self, value: T, count: int, selected: bool,
            facet_has_selected_values: bool
    ):
        self.value = value
        self.count = count
        self.selected = selected
        self._facet_has_selected_values = facet_has_selected_values

    @property
    def count_text(self) -> str:
        if self._facet_has_selected_values and not self.selected:
            return f'+{self.count}'
        return f'{self.count}'


class AttrFacet(t.Generic[T]):
    def __init__(self, attr_id: int):
        self.attr_id = attr_id
        self.values: t.List[AttrFacetValue[T]] = []
        self.selected_values: t.List[AttrFacetValue[T]] = []
        self.all_values: t.List[AttrFacetValue[T]] = []
        self._values_map: t.Dict[T, AttrFacetValue[T]] = {}

    def add_value(self, facet_value: AttrFacetValue[T]) -> None:
        if facet_value.selected:
            self.selected_values.append(facet_value)
        else:
            self.values.append(facet_value)
        self.all_values.append(facet_value)
        self._values_map[facet_value.value] = facet_value

    def get_value(self, value: T) -> t.Optional[AttrFacetValue[T]]:
        return self._values_map.get(value)


class AttrFacetFilterResult(BaseFilterResult, t.Generic[T]):
    def __init__(self, name: str, alias: str):
        super().__init__(name, alias)
        self.facets: t.Dict[int, AttrFacet[T]] = {}

    def add_attr_value(
            self, attr_id: int, facet_value: AttrFacetValue[T]
    ) -> None:
        facet = self.facets.get(attr_id)
        if facet is None:
            facet = AttrFacet(attr_id)
            self.facets[attr_id] = facet
        facet.add_value(facet_value)

    def get_facet(self, attr_id: int) -> t.Optional[AttrFacet[T]]:
        return self.facets.get(attr_id)


class AttrRangeFacet:
    def __init__(
        self,
        attr_id: int,
        count: int,
        selected: bool,
        min_: TMinValue = None,
        max_: TMaxValue = None,
    ):
        self.attr_id = attr_id
        self.count = count
        self.selected = selected
        self.min = min_
        self.max = max_


class AttrRangeFacetFilterResult(BaseFilterResult):
    def __init__(self, name: str, alias: str):
        super().__init__(name, alias)
        self.facets: t.Dict[int, AttrRangeFacet] = {}

    def add_facet(self, facet: AttrRangeFacet) -> None:
        self.facets[facet.attr_id] = facet

    def get_facet(self, attr_id: int) -> t.Optional[AttrRangeFacet]:
        return self.facets.get(attr_id)



import struct
import typing


def merge_attr_value_int(attr_id: int, value_id: int) -> int:
    return (attr_id << 32) | (value_id & 0xffff_ffff)


def split_attr_value_int(merged_attr: int) -> typing.Tuple[int, int]:
    return merged_attr >> 32, merged_attr & 0xffff_ffff


def merge_attr_value_bool(attr_id: int, value: bool) -> int:
    return (attr_id << 1) | value


def split_attr_value_bool(merged_attr: int) -> typing.Tuple[int, bool]:
    return merged_attr >> 1, bool(merged_attr & 1)


def merge_attr_value_float(attr_id: int, value: float) -> int:
    return (attr_id << 32) | struct.unpack('=I', struct.pack('=f', value))[0]



from abc import ABC
import typing as t

from elasticmagic import Bool, Range, Term, Terms
from elasticmagic import SearchQuery
from elasticmagic import types
from elasticmagic.expression import Expression, FieldOperators
from elasticmagic.ext.queryfilter.queryfilter import BaseFilter

from elasticmagic.ext.queryfilter.codec import FloatCodec
from elasticmagic.ext.queryfilter.codec import IntCodec

from .util import merge_attr_value_bool
from .util import merge_attr_value_float
from .util import merge_attr_value_int


ParamValues = t.Dict[str, t.List[str]]
Params = t.Dict[str, ParamValues]

T = t.TypeVar('T')

float_codec = FloatCodec()
int_codec = IntCodec()


class BaseAttrSimpleFilter(ABC, BaseFilter, t.Generic[T]):
    def __init__(self, name: str, field: FieldOperators, alias: str = None):
        super().__init__(name, alias=alias)
        self.field = field

    def _iter_attr_values(
            self, params: Params
    ) -> t.Generator[t.Tuple[int, t.Dict[str, t.List[str]]], None, None]:
        for k, w in params.items():
            if not k.startswith(self.alias):
                continue
            try:
                attr_id = int_codec.decode(
                    k[len(self.alias):], es_type=types.Integer
                )
            except ValueError:
                continue
            yield (attr_id, w)

    def _apply_filter(
        self, search_query: SearchQuery, params: Params
    ) -> SearchQuery:
        for attr_id, w in self._iter_attr_values(params):
            expr = self._get_filter_expression(attr_id, w)
            if not expr:
                continue
            search_query = self._apply_filter_expression(
                search_query, expr, attr_id
            )
        return search_query

    def _apply_filter_expression(
        self, search_query: SearchQuery, expr: Expression, attr_id: int
    ) -> SearchQuery:
        return search_query.filter(expr)

    def _get_filter_expression(
            self, attr_id: int, values: ParamValues
    ) -> t.Optional[Expression]:
        raise NotImplementedError  # pragma: no cover

    @staticmethod
    def _parse_value(v: str) -> T:
        raise NotImplementedError  # pragma: no cover

    @classmethod
    def _parse_values(cls, values: ParamValues, op: str) -> t.List[T]:
        w = []
        for v in values.get(op, []):
            try:
                w.append(cls._parse_value(v))
            except ValueError:
                continue
        return w


class AttrIntSimpleFilter(BaseAttrSimpleFilter[int]):
    @staticmethod
    def _parse_value(v: str) -> int:
        return int_codec.decode(v, es_type=types.Integer)

    def _get_filter_expression(
        self, attr_id: int, values
    ) -> t.Optional[Expression]:
        w = [
            merge_attr_value_int(attr_id, v)
            for v in self._parse_values(values, 'exact')
        ]
        if not w:
            return None
        if len(w) == 1:
            return Term(self.field, w[0])
        return Terms(self.field, w)


class AttrBoolSimpleFilter(BaseAttrSimpleFilter[bool]):
    @staticmethod
    def _parse_value(v: str) -> bool:
        if v == 'true' or v == 'True':
            return True
        if v == 'false' or v == 'False':
            return False
        raise ValueError(f'Cannot parse boolean value: {v}')

    def _get_filter_expression(
            self, attr_id: int, values: ParamValues
    ) -> t.Optional[Expression]:
        w = [
            merge_attr_value_bool(attr_id, v)
            for v in self._parse_values(values, 'exact')
        ]
        if not w:
            return None
        if len(w) == 1:
            return Term(self.field, w[0])
        return Terms(self.field, w)


#
#             -Inf                 +0.0
#    0x{attr_id}_ff800000 0x{attr_id}_00000000
#                       | |
#                    *********
#                 **           **
#                *               *
#    negative   *                 *    positive
#    floats     *                 *    floats
#             ⤹ *                * ⤸
#                 **           **
#                    *********
#                       | |
#    0x{attr_id}_80000000 0x{attr_id}_7f800000
#             -0.0                 +Inf
#
class AttrRangeSimpleFilter(BaseAttrSimpleFilter[float]):
    @staticmethod
    def _parse_value(v: str) -> float:
        return float_codec.decode(v)

    @classmethod
    def _parse_last_value(
            cls, values: ParamValues, op: str
    ) -> t.Optional[float]:
        parsed_values = cls._parse_values(values, op)
        if not parsed_values:
            return None
        return parsed_values[-1]

    @staticmethod
    def _plus_zero(attr_id: int) -> int:
        return merge_attr_value_float(attr_id, 0.0)

    @staticmethod
    def _minus_zero(attr_id: int) -> int:
        return merge_attr_value_float(attr_id, -0.0)

    @staticmethod
    def _plus_inf(attr_id: int) -> int:
        return merge_attr_value_float(attr_id, float('inf'))

    @staticmethod
    def _minus_inf(attr_id: int) -> int:
        return merge_attr_value_float(attr_id, float('-inf'))

    def _get_filter_expression(
        self, attr_id: int, values: ParamValues
    ) -> t.Optional[Expression]:
        gte = self._parse_last_value(values, 'gte')
        gte_value = None
        if gte is not None:
            gte_value = merge_attr_value_float(attr_id, gte)

        lte = self._parse_last_value(values, 'lte')
        lte_value = None
        if lte is not None:
            lte_value = merge_attr_value_float(attr_id, lte)

        if gte is not None and lte is not None:
            if gte >= 0.0 and lte >= 0.0:
                return Range(self.field, gte=gte_value, lte=lte_value)
            elif gte < 0.0 and lte < 0.0:
                return Range(self.field, gte=lte_value, lte=gte_value)
            elif gte < 0.0 and lte >= 0:
                return Bool.should(
                    Range(
                        self.field,
                        gte=self._minus_zero(attr_id), lte=gte_value
                    ),
                    Range(
                        self.field, gte=self._plus_zero(attr_id), lte=lte_value
                    ),
                )
            else:
                return Bool.must(
                    Range(
                        self.field, gte=gte_value, lte=self._plus_inf(attr_id)
                    ),
                    Range(
                        self.field, gte=lte_value, lte=self._minus_inf(attr_id)
                    ),
                )

        if gte is not None:
            if gte >= 0.0:
                return Range(
                    self.field, gte=gte_value, lte=self._plus_inf(attr_id)
                )
            else:
                return Bool.should(
                    Range(
                        self.field,
                        gte=self._minus_zero(attr_id), lte=gte_value
                    ),
                    Range(
                        self.field,
                        gte=self._plus_zero(attr_id),
                        lte=self._plus_inf(attr_id)
                    ),
                )

        if lte is not None:
            if lte < 0.0:
                return Range(
                    self.field, gte=lte_value, lte=self._minus_inf(attr_id)
                )
            else:
                return Bool.should(
                    Range(
                        self.field, gte=self._plus_zero(attr_id), lte=lte_value
                    ),
                    Range(
                        self.field,
                        gte=self._minus_zero(attr_id),
                        lte=self._minus_inf(attr_id)
                    ),
                )

        return None



import typing as t

from elasticmagic import agg
from elasticmagic import Bool
from elasticmagic import Range
from elasticmagic import Script
from elasticmagic import SearchQuery
from elasticmagic.agg import SingleValueMetricsAggResult
from elasticmagic.expression import Expression
from elasticmagic.expression import FieldOperators
from elasticmagic.result import SearchResult

from .facet_result import AttrFacetFilterResult, TMaxValue, TMinValue
from .facet_result import AttrRangeFacet
from .facet_result import AttrRangeFacetFilterResult
from .facet_result import AttrFacetValue
from .simple import AttrBoolSimpleFilter
from .simple import AttrRangeSimpleFilter
from .simple import AttrIntSimpleFilter
from .simple import BaseAttrSimpleFilter
from .simple import Params
from .util import merge_attr_value_bool
from .util import merge_attr_value_int
from .util import split_attr_value_bool
from .util import split_attr_value_int


T = t.TypeVar('T')


RANGE_ATTR_SCRIPT = '''
int attrsLen = doc[params.field].size();
List values = doc[params.field];

int[] attrIds = new int[attrsLen];
for (int i = 0; i < attrsLen; i++) {
    attrIds[i] = (int) (values[i] >>> 32);
}
return attrIds;
'''

RANGE_ATTR_MINMAX_MAP_SCRIPT = '''
for (v in doc[params.field]) {
    String attr_id = (v >>> 32).toString();
    float float_val = Float.intBitsToFloat((int) v);
    float[] min_max = state.get(attr_id);
    if (min_max == null) {
        state[attr_id] = new float[] {float_val, float_val};
        continue;
    }
    if (float_val < min_max[0]){
        min_max[0] = float_val;
    }
    if (float_val > min_max[1]){
        min_max[1] = float_val;
    }
}
'''

RANGE_ATTR_MINMAX_REDUCE_SCRIPT = '''
Map reduced = new HashMap();
for (state in states) {
    for (entry in state.entrySet()) {
        String attr_id = entry.getKey();
        float[] min_max = entry.getValue();
        if (!reduced.containsKey(attr_id)) {
            reduced[attr_id] = min_max;
            continue;
        }
        if (min_max[0] < reduced[attr_id][0]) {
            reduced[attr_id][0] = min_max[0];
        }
        if (min_max[1] > reduced[attr_id][1]) {
            reduced[attr_id][1] = min_max[1];
        }
    }
}
return reduced;
'''

RANGE_ATTR_MINMAX_COMBINE_SCRIPT = 'return state;'


def _parse_attr_id_from_agg_name(agg_name: str) -> t.Optional[int]:
    try:
        return int(agg_name.rpartition(':')[2])
    except ValueError:
        return None


class BaseAttrFacetFilter(BaseAttrSimpleFilter, t.Generic[T]):
    full_agg_size: int
    single_agg_size: int

    _result_cls: t.Type[AttrFacetFilterResult[T]]

    _facet_value_cls: t.Type[AttrFacetValue[T]]

    _attr_id_meta_key: str

    def _split_bucket_key(self, key: int) -> t.Tuple[int, T]:
        raise NotImplementedError  # pragma: no cover

    def _include_attrs_values(
            self, attr_ids: t.Iterable[int]
    ) -> t.Dict[int, t.List[int]]:
        raise NotImplementedError  # pragma: no cover

    def _apply_filter_expression(
            self, search_query: SearchQuery, expr: Expression, attr_id: int
    ) -> None:
        return search_query.post_filter(
            expr,
            meta={
                'tags': {self.name, f'{self.alias}:{attr_id}'},
                self._attr_id_meta_key: attr_id
            }
        )

    @property
    def _agg_name(self) -> str:
        return f'{self.qf._name}.{self.name}'

    @property
    def _filter_agg_name(self) -> str:
        return f'{self.qf._name}.{self.name}.filter'

    def _apply_agg(self, search_query: SearchQuery) -> SearchQuery:
        aggs = {}

        exclude_tags = {self.qf._name}
        filters = self._get_agg_filters(
            search_query.get_context().iter_post_filters_with_meta(),
            exclude_tags
        )

        full_terms_agg = agg.Terms(
            self.field, size=self.full_agg_size
        )
        if filters:
            aggs[self._filter_agg_name] = agg.Filter(
                Bool.must(*filters),
                aggs={self._agg_name: full_terms_agg}
            )
        else:
            aggs[self._agg_name] = full_terms_agg

        post_filters = list(
            search_query.get_context().iter_post_filters_with_meta()
        )
        selected_attr_ids = []
        for filt, meta in post_filters:
            if not meta:
                continue
            attr_id = meta.get(self._attr_id_meta_key)
            if attr_id is None:
                continue
            selected_attr_ids.append(attr_id)
        include_attrs_values = self._include_attrs_values(selected_attr_ids)
        for attr_id in selected_attr_ids:
            attr_aggs = {
                f'{self._agg_name}:{attr_id}': agg.Terms(
                    self.field,
                    size=self.single_agg_size,
                    include=include_attrs_values.get(attr_id),
                )
            }
            filters = [
                f for f, m in post_filters
                if m.get(self._attr_id_meta_key) != attr_id
            ]
            if filters:
                aggs[f'{self._filter_agg_name}:{attr_id}'] = agg.Filter(
                    Bool.must(*filters),
                    aggs=attr_aggs
                )
            else:
                aggs.update(attr_aggs)

        return search_query.aggs(aggs)

    def _process_result(
            self, result: SearchResult, params: Params
    ) -> AttrFacetFilterResult[T]:
        facet_result = self._result_cls(self.name, self.alias)

        selected_attr_values = {}
        for selected_attr_id, w in self._iter_attr_values(params):
            selected_attr_values[selected_attr_id] = set(
                self._parse_values(w, 'exact')
            )

        processed_attr_ids = set()
        for agg_name, attr_agg in result.aggregations.items():
            if agg_name.startswith(f'{self._filter_agg_name}:'):
                attr_id = _parse_attr_id_from_agg_name(agg_name)
                if attr_id is not None:
                    agg_name = f'{self._agg_name}:{attr_id}'
                    attr_agg = attr_agg.get_aggregation(agg_name)

            if not agg_name.startswith(f'{self._agg_name}:'):
                continue

            attr_id = _parse_attr_id_from_agg_name(agg_name)
            if attr_id is None:
                continue
            selected_values = selected_attr_values.get(attr_id) or set()
            processed_attr_ids.add(attr_id)
            for bucket in attr_agg.buckets:
                found_attr_id, value_id = self._split_bucket_key(bucket.key)
                if found_attr_id != attr_id:
                    continue
                fv = self._facet_value_cls(
                    value_id,
                    bucket.doc_count,
                    value_id in selected_values,
                    bool(selected_values),
                )
                facet_result.add_attr_value(attr_id, fv)

        main_agg = result.get_aggregation(self._agg_name)
        if main_agg is None:
            main_agg = result.get_aggregation(self._filter_agg_name) \
                .get_aggregation(self._agg_name)
        for bucket in main_agg.buckets:
            attr_id, value_id = self._split_bucket_key(bucket.key)
            if attr_id in processed_attr_ids:
                continue
            fv = self._facet_value_cls(
                value_id, bucket.doc_count, False, False
            )
            facet_result.add_attr_value(attr_id, fv)

        return facet_result


AttrsValuesGetter = t.Callable[[t.Iterable[int]], t.Dict[int, t.List[int]]]


class AttrIntFacetFilter(AttrIntSimpleFilter, BaseAttrFacetFilter[int]):
    _result_cls = AttrFacetFilterResult[int]
    _facet_value_cls = AttrFacetValue[int]

    _attr_id_meta_key = 'int_attr_id'

    def __init__(
            self, name: str, field: FieldOperators,
            alias: t.Optional[str] = None,
            full_agg_size: int = 10_000,
            single_agg_size: int = 100,
            attrs_values_getter: t.Optional[AttrsValuesGetter] = None,
    ):
        super().__init__(name, field, alias=alias)
        self.full_agg_size = full_agg_size
        self.single_agg_size = single_agg_size
        self._attrs_values_getter = attrs_values_getter

    def _split_bucket_key(self, key: int) -> t.Tuple[int, int]:
        return split_attr_value_int(key)

    def _include_attrs_values(
            self, attr_ids: t.Iterable[int]
    ) -> t.Dict[int, t.List[int]]:
        if self._attrs_values_getter is None:
            return {}
        attrs_values = {}
        for attr_id, values in self._attrs_values_getter(attr_ids).items():
            attrs_values[attr_id] = [
                merge_attr_value_int(attr_id, v) for v in values
            ]
        return attrs_values


class AttrBoolFacetFilter(AttrBoolSimpleFilter, BaseAttrFacetFilter[bool]):
    _result_cls = AttrFacetFilterResult[bool]
    _facet_value_cls = AttrFacetValue[bool]

    _attr_id_meta_key = 'bool_attr_id'

    def __init__(
            self, name: str, field: FieldOperators,
            alias: t.Optional[str] = None,
            full_agg_size: int = 100,
            single_agg_size: int = 2,
    ):
        super().__init__(name, field, alias=alias)
        self.full_agg_size = full_agg_size
        self.single_agg_size = single_agg_size

    def _split_bucket_key(self, key: int) -> t.Tuple[int, bool]:
        return split_attr_value_bool(key)

    def _include_attrs_values(
            self, attr_ids: t.Iterable[int]
    ) -> t.Dict[int, t.List[int]]:
        attrs_values = {}
        for attr_id in attr_ids:
            attrs_values[attr_id] = [
                merge_attr_value_bool(attr_id, False),
                merge_attr_value_bool(attr_id, True),
            ]
        return attrs_values


class AttrRangeFacetFilter(AttrRangeSimpleFilter):
    _attr_id_meta_key = 'float_attr_id'

    def __init__(
            self,
            name: str,
            field: FieldOperators,
            alias: t.Optional[str] = None,
            compute_min_max: bool = False,
    ):
        super().__init__(name, field, alias=alias)
        self._compute_min_max = compute_min_max

    def _apply_filter_expression(
            self, search_query: SearchQuery, expr: Expression, attr_id: int
    ) -> None:
        return search_query.post_filter(
            expr,
            meta={
                'tags': {self.name, f'{self.alias}:{attr_id}'},
                self._attr_id_meta_key: attr_id
            }
        )

    def _agg_name(self, attr_id: t.Optional[int] = None) -> str:
        agg_name = f'{self.qf._name}.{self.name}'
        if attr_id is not None:
            return f'{agg_name}:{attr_id}'
        return agg_name

    def _filter_agg_name(self) -> str:
        return f'{self._agg_name()}.filter'

    def _min_max_agg_name(self) -> str:
        return f'{self._agg_name()}.min_max'

    def _filter_min_max_agg_name(self) -> str:
        return f'{self._min_max_agg_name()}.filter'

    def _process_min_max_agg_result(
            self,
            attr_id: int,
            min_max_agg: SingleValueMetricsAggResult,
    ) -> t.Tuple[TMinValue, TMaxValue]:

        min_ = None
        max_ = None

        if (
            self._compute_min_max
            and min_max_agg.value
            and str(attr_id) in min_max_agg.value
        ):
            min_, max_ = min_max_agg.value[str(attr_id)]

        return min_, max_

    def _apply_agg(self, search_query: SearchQuery) -> SearchQuery:
        aggs = {}

        post_filters_with_meta = list(
            search_query.get_context().iter_post_filters_with_meta()
        )
        exclude_tags = {self.qf._name}
        agg_filters = self._get_agg_filters(
            post_filters_with_meta,
            exclude_tags,
        )

        full_terms_agg = agg.Terms(
            script=Script(
                RANGE_ATTR_SCRIPT,
                lang='painless',
                params={
                    'field': self.field,
                }
            ),
            size=100
        )
        if agg_filters:
            aggs[self._filter_agg_name()] = agg.Filter(
                Bool.must(*agg_filters),
                aggs={self._agg_name(): full_terms_agg}
            )
        else:
            aggs[self._agg_name()] = full_terms_agg

        for filt, meta in post_filters_with_meta:
            if not meta:
                continue
            selected_attr_id = meta.get(self._attr_id_meta_key)
            if selected_attr_id is None:
                continue

            filters = [
                f for f, m in post_filters_with_meta
                if m.get(self._attr_id_meta_key) != selected_attr_id
            ]
            filters.append(
                Range(
                    self.field,
                    gte=merge_attr_value_int(selected_attr_id, 0),
                    lte=merge_attr_value_int(selected_attr_id, 0xffff_ffff)
                )
            )
            aggs[self._agg_name(selected_attr_id)] = agg.Filter(
                Bool.must(*filters)
            )

        if self._compute_min_max:
            min_max_agg = agg.ScriptedMetric(
                map_script=RANGE_ATTR_MINMAX_MAP_SCRIPT,
                reduce_script=RANGE_ATTR_MINMAX_REDUCE_SCRIPT,
                combine_script=RANGE_ATTR_MINMAX_COMBINE_SCRIPT,
                params={
                    'field': self.field,
                },
            )
            min_max_filters = [
                f for f, m in post_filters_with_meta
                if m.get(self._attr_id_meta_key) is None
                and not m.get('tags', set()).intersection(exclude_tags)
            ]
            if min_max_filters:
                aggs[self._filter_min_max_agg_name()] = agg.Filter(
                    Bool.must(*min_max_filters),
                    aggs={self._min_max_agg_name(): min_max_agg}
                )
            else:
                aggs[self._min_max_agg_name()] = min_max_agg

        return search_query.aggs(aggs)

    def _process_result(
            self, result: SearchResult, params: t.Dict
    ) -> AttrRangeFacetFilterResult:
        facet_result = AttrRangeFacetFilterResult(self.name, self.alias)

        selected_attr_ids = set()
        for selected_attr_id, w in self._iter_attr_values(params):
            if (
                self._parse_last_value(w, 'gte') is not None or
                self._parse_last_value(w, 'lte') is not None
            ):
                selected_attr_ids.add(selected_attr_id)

        main_agg = result.get_aggregation(self._agg_name())
        if main_agg is None:
            main_agg = result.get_aggregation(self._filter_agg_name()) \
                .get_aggregation(self._agg_name())

        min_max_agg = None
        if self._compute_min_max:
            min_max_agg = result.get_aggregation(self._min_max_agg_name())
            if not min_max_agg:
                min_max_agg = (
                    result
                    .get_aggregation(self._filter_min_max_agg_name())
                    .get_aggregation(self._min_max_agg_name())
                )

        for bucket in main_agg.buckets:
            attr_id = int(bucket.key)
            min_, max_ = self._process_min_max_agg_result(
                attr_id, min_max_agg
            )
            facet_result.add_facet(
                AttrRangeFacet(
                    attr_id=attr_id,
                    count=bucket.doc_count,
                    selected=attr_id in selected_attr_ids,
                    min_=min_,
                    max_=max_,
                )
            )

        for selected_attr_id in selected_attr_ids:
            selected_agg = result.get_aggregation(
                self._agg_name(selected_attr_id)
            )
            min_, max_ = self._process_min_max_agg_result(
                selected_attr_id, min_max_agg
            )
            facet_result.add_facet(
                AttrRangeFacet(
                    attr_id=selected_attr_id,
                    count=selected_agg.doc_count,
                    selected=True,
                    min_=min_,
                    max_=max_,
                )
            )

        return facet_result



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['elasticmagic_qf_attrs']

package_data = \
{'': ['*']}

install_requires = \
['elasticmagic>=0.1.0-beta.1,<0.2.0']

setup_kwargs = {
    'name': 'elasticmagic-qf-attrs',
    'version': '0.1.4',
    'description': '',
    'long_description': '[![Test and build](https://github.com/anti-social/elasticmagic-qf-attrs/actions/workflows/python.yaml/badge.svg)](https://github.com/anti-social/elasticmagic-qf-attrs/actions/workflows/python.yaml)\n\nElasticmagic query filters for attributes\n=========================================\n\nLibrary to store, filter and build facets for custom attributes\n\nThe problem\n-----------\n\nEach attribute pair can be stored in the index as a nested document.\nWe can use following mapping for that:\n\n```yaml\nattrs:\n  type: nested\n  properties:\n    attr_id:\n      type: integer\n    # usually only one of the next fields should be populated\n    value_int:\n      type: integer\n    value_bool:\n      type: boolean\n    value_float:\n      type: float\n```\n\nThis makes it possible to filter documents by an attribute id and its value\n(for example we want to find all the documents with `attr_id = 1234` and `value = 5678`):\n\n```yaml\nquery:\n  bool:\n    filter:\n    - nested:\n        path: attrs\n        query:\n          bool:\n            must:\n            - term:\n                attrs.attr_id: 1234\n            - term:\n                attrs.value_int: 5678\n```\n\nIt is also possible to build a facets for all attributes at once:\n\n```yaml\naggs:\n  attrs_nested:\n    nested:\n      path: attrs\n    aggs:\n      attrs:\n        terms:\n          field: attrs.attr_id\n        aggs:\n          values:\n            field: attrs.value_int\n```\n\nor for a single attribute:\n\n```yaml\naggs:\n  attrs_nested:\n    nested:\n      path: attrs\n    aggs:\n      attr_1234:\n        filter:\n          term:\n            attrs.attr_id: 1234\n        aggs:\n          values:\n            field: attrs.value_int\n```\n\nBut nested documents have some drawbacks. Every nested document is stored\nin the index as different document. For instance, next document will be stored\nas 5 lucene documents:\n\n```yaml\nname: "I\'m a document with nested attributes"\nattrs:\n- attr_id: 1\n  value_int: 42\n- attr_id: 2\n  value_int: 43\n- attr_id: 3\n  value_bool: true\n- attr_id: 4\n  value_float: 99.9\n```\n\nNested queries are slow by itself:\n   \n> In particular, joins should be avoided. nested can make queries several times\nslower and parent-child relations can make queries hundreds of times slower.\n> - https://www.elastic.co/guide/en/elasticsearch/reference/master/tune-for-search-speed.html#_document_modeling\n\nBut what is worse regular queries are also slower when there are nested documents\nin the index. It is because of all the fields of main documents becomes sparse.\nThis in turn degrades performance of all filters and accesses to doc_values.\n\nThe solution\n------------\n\nThe idea is to encode pair of an attribute id and a corresponding value into\na single value. If our attribute ids are 32-bit integers and all value types\nalso fit into 32 bits we can store them as a single 64-bit value.\n\nSo our mapping can be:\n\n```yaml\nattrs:\n  type: object\n  properties:\n    int:\n      type: long\n    bool:\n      type: long\n    float:\n      type: long\n```\n\nDocument with encoded attributes:\n\n```yaml\nname: "I\'m a document with packed attributes"\nattrs:\n# (1 << 32) | 42\n- int: 0x1_0000002a\n# (2 << 32) | 43\n- int: 0x2_0000002b\n# (3 << 1) | 1\n- bool: 0x7\n# (4 << 32) | {integer representation of 99.9}\n# (4 << 32) | struct.unpack(\'=I\', struct.pack(\'=f\', 99.9))[0]\n- float: 0x4_42c7cccd\n```\n\nNow with a bit of bit magic we can emulate nested queries.\n\nFiltering by attribute id `1234` with value `5678`:\n\n```yaml\nquery:\n  bool:\n    filter:\n    - term:\n        attrs.int: 0x4d2_0000162e\n```\n\nBuilding facet for all attribute values:\n\n```yaml\naggs:\n  attrs_int:\n    terms:\n      field: attrs.int\n      # specify big enough aggregation size\n      # so all flat attrite values should fit\n      size: 10000\n```\n\nOne more step that client should do is to decode and group values by\nattribute id.\n\nHow to use it in python\n-----------------------\n\n```python\nfrom elasticsearch import Elasticsearch\n\nfrom elasticmagic import Cluster, Document, Field\nfrom elasticmagic.types import List, Long\nfrom elasticmagic.ext.queryfilter import QueryFilter\n\nfrom elasticmagic_qf_attrs import AttrBoolFacetFilter\nfrom elasticmagic_qf_attrs import AttrIntFacetFilter\nfrom elasticmagic_qf_attrs import AttrRangeFacetFilter\nfrom elasticmagic_qf_attrs.util import merge_attr_value_bool\nfrom elasticmagic_qf_attrs.util import merge_attr_value_float\nfrom elasticmagic_qf_attrs.util import merge_attr_value_int\n\n# Specify document\nclass AttrsDocument(Document):\n    __doc_type__ = \'attrs\'\n\n    ints = Field(List(Long))\n    bools = Field(List(Long))\n    floats = Field(List(Long))\n\n# Create an index\nindex_name = \'test-attrs\'\nclient = Elasticsearch()\nclient.indices.create(index=index_name)\ncluster = Cluster(client)\nindex = cluster.get_index(index_name)\nindex.put_mapping(AttrsDocument)\n\n# Index example document\nindex.add([\n    AttrsDocument(\n        ints=[\n            merge_attr_value_int(1, 42),\n            merge_attr_value_int(2, 43),\n        ],\n        bools=[merge_attr_value_bool(3, True)],\n        floats=[merge_attr_value_float(4, 99.9)],\n    ),\n], refresh=True)\n\n# Define a query filter\nclass AttrsQueryFilter(QueryFilter):\n    ints = AttrIntFacetFilter(AttrsDocument.ints, alias=\'a\')\n    bools = AttrBoolFacetFilter(AttrsDocument.bools, alias=\'a\')\n    ranges = AttrRangeFacetFilter(AttrsDocument.floats, alias=\'a\')\n\n# Now we can build facets\nqf = AttrsQueryFilter()\nsq = index.search_query()\nsq = qf.apply(sq, {})\nres = sq.get_result()\nassert res.total == 1\nqf_res = qf.process_result(res)\n\n# And finally lets print results\nfor attr_id, facet in qf_res.ints.facets.items():\n    print(f\'> {attr_id}:\')\n    for facet_value in facet.all_values:\n        print(f\'  {facet_value.value}: ({facet_value.count_text})\')\n\nfor attr_id, facet in qf_res.bools.facets.items():\n    print(f\'> {attr_id}:\')\n    for facet_value in facet.all_values:\n        print(f\'  {facet_value.value}: ({facet_value.count_text})\')\n\nfor attr_id, facet in qf_res.ranges.facets.items():\n    print(f\'> {attr_id}: ({facet.count})\')\n\n# Also we can filter documents:\nqf = AttrsQueryFilter()\nsq = index.search_query()\nsq = qf.apply(\n    sq,\n    {\n        \'a1\': \'42\',\n        \'a3\': \'true\',\n        \'a4__lte\': \'100\',\n    }\n)\nres = sq.get_result()\nassert res.total == 1\n\nqf = AttrsQueryFilter()\nsq = index.search_query()\nsq = qf.apply(\n    sq,\n    {\n        \'a4__gte\': \'100\',\n    }\n)\nres = sq.get_result()\nassert res.total == 0\n```\n\nThis script should print:\n\n```\n> 1:\n  42: (1)\n> 2:\n  43: (1)\n> 3:\n  True: (1)\n> 4: (1)\n```\n',
    'author': 'Alexander Koval',
    'author_email': 'kovalidis@gmail.com',
    'maintainer': 'Alexander Koval',
    'maintainer_email': 'kovalidis@gmail.com',
    'url': 'https://github.com/anti-social/elasticmagic-qf-attrs',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
